import os
from types import SimpleNamespace
from typing import Dict

from . import globs


class StaticRegistry:

    static_file = staticmethod(globs.static_file)

    class Registered(SimpleNamespace):
        folder: str
        client_apps: set

    mounted: Dict[str, Registered] = {}

    def register(self, base_url, folder, app):
        folder_apps = self.mounted.get(base_url)
        if folder_apps:
            if os.path.samefile(folder_apps.folder, folder):
                raise KeyError(
                    f'URL already in use: {base_url} => path:'
                    f'{folder_apps.folder}'
                )
            folder_apps.client_apps.add(app)
        self.mounted[base_url] = SimpleNamespace(
            folder=folder, client_apps={app}
        )

    def unregister(self, base_url, app):
        reg = self.mounted.get(base_url)
        if not reg:
            return
        reg.client_apps.difference_update({app})
        if not reg.client_apps:
            self.mounted.pop(base_url)

    def get_registered(self, base_url, folder):
        folder_apps = self.mounted.get(base_url)
        if not folder_apps:
            return None
        if os.path.samefile(folder_apps.folder, folder):
            return folder_apps

    def make_rule_and_handler(self, static_base_url: str, folder: str, client_app):
        if not os.path.exists(folder):
            return None, None
        registered = self.get_registered(static_base_url, folder)
        if registered:
            registered.client_apps.add(client_app)
            return None, None

        self.register(static_base_url, folder, client_app)
        static_ver_rex = r'<re((_\d+(\.\d+){2}/)?)>'
        if static_base_url.endswith('static'):
            static_base_url = f'{static_base_url}/{static_ver_rex}'
        elif '/static/' in static_base_url:
            static_base_url = static_base_url.replace('/static/', f'/static/{static_ver_rex}', 1)
        if not static_base_url.endswith(('/', '/)?)>')):
            static_base_url = f'{static_base_url}/'
        rule = f'{static_base_url}<fp.path()>'
        h = self.make_static_handler(folder)
        return rule, h

    @classmethod
    def make_static_handler(cls, folder):
        response = globs.response

        def serve_static(fp):
            response.headers.setdefault("Pragma", "cache")
            response.headers.setdefault("Cache-Control", "private")
            return cls.static_file(fp, root=folder)
        return serve_static

    def __contains__(self, base_url_folder_tuple):
        if isinstance(base_url_folder_tuple, str):
            raise TypeError(
                f'A pair like [url, folder] is required, '
                f'got string: {base_url_folder_tuple}'
            )
        base_url, folder = base_url_folder_tuple
        folder_apps = self.mounted.get(base_url)
        if not folder_apps:
            return False
        return os.path.samefile(folder_apps.folder, folder)


static_registry = StaticRegistry()



import re
import sys
import os
import click
import importlib
import traceback
from typing import Dict

from .globs import app as om_app
from .loggers import error_logger, get_error_snapshot
from .utils import module2filename
from .static_registry import static_registry

from .core_events import core_event_bus, CoreEvents


def _clear_modules(module_name):
    # all files/submodules
    names = [
        name
        for name in sys.modules
        if (name + ".").startswith(module_name + ".")
    ]
    for name in names:
        del sys.modules[name]


class Reloader:

    modules = {}
    errors = {}
    apps_data: Dict[str, list] = {}
    current_import_app: str

    @staticmethod
    def get_apps_folder():
        return os.environ["WEBSAW_APPS_FOLDER"]

    @classmethod
    def register_app_data(cls, app_data):
        app_lst = cls.apps_data.setdefault(cls.current_import_app, [])
        app_lst.append(app_data)

    @classmethod
    def install_reloader_hook(cls):
        core_event_bus.on(CoreEvents.RELOAD_APPS, cls.reimport_apps)

    @classmethod
    def expand_apps_to_reload(cls, *app_names):
        ret = set()
        for app_name in app_names:
            ret.add(app_name)
            module_name = f"apps.{app_name}"
            app = getattr(sys.modules.get(module_name), 'app', None)
            get_all_clients = getattr(app, 'get_all_clients', None)
            if not get_all_clients:
                continue
            clients = get_all_clients()
            for m_name in (capp._module_name for capp in clients):
                if not re.match(r'apps\.\w+', m_name):
                    continue
                ret.add(m_name.split('.')[1])
        return ret

    @classmethod
    def clear_routes(cls, app_name=''):
        app_name.strip('/')
        root = f'/{app_name}' if app_name else ''
        routes = f'{root}/*'
        om_app.router.remove(routes)
        if root:
            om_app.router.remove(root)
            root_slash = f'{root}/'
            for app_data in cls.apps_data.pop(app_name, []):
                for route in app_data.routes:
                    if route.rule == root or route.rule.startswith(root_slash):
                        # already removed
                        continue
                    om_app.router.remove(route)
                static_registry.unregister(app_data.static_base_url, app_name)

    @classmethod
    def reimport_apps(cls, *apps):
        if not apps:
            return
        apps = cls.expand_apps_to_reload(*apps)
        for app in apps:
            _clear_modules(f"apps.{app}")
            cls.clear_routes(app)
        for app in apps:
            cls.import_app(app, clear_before_import=False)

    @classmethod
    def import_apps(cls):
        """Import or reimport modules"""
        cls.clear_routes()
        folder = cls.get_apps_folder()
        # if first time reload dummy top module
        if not cls.modules:
            path = os.path.join(folder, "__init__.py")
            loader = importlib.machinery.SourceFileLoader("apps", path)
            loader.load_module()
        # Then load all the apps as submodules
        for app_name in os.listdir(folder):
            cls.import_app(app_name, clear_before_import=False)

    @classmethod
    def import_app(cls, app_name, clear_before_import=True):
        folder = cls.get_apps_folder()
        path = os.path.join(folder, app_name)
        init = os.path.join(path, "__init__.py")

        if not (
            os.path.isdir(path)
            and not path.endswith("__")
            and os.path.exists(init)
        ):
            return

        if clear_before_import:
            cls.clear_routes(app_name)
        cls.current_import_app = app_name

        module_name = f"apps.{app_name}"
        is_loaded = module_name in sys.modules
        try:
            module = cls.modules.get(app_name)
            if module is None:
                if is_loaded:
                    click.secho(f"[X] already loaded {app_name}       ", fg="green")
                else:
                    click.echo(f"[ ] loading {app_name} ...")
            else:
                if is_loaded:
                    click.secho(f"[X] already reloaded {app_name}       ", fg="green")
                else:
                    click.echo(f"[ ] reloading {app_name} ...")
                    # forget the module
                    del cls.modules[app_name]

            if not is_loaded:
                module = importlib.machinery.SourceFileLoader(
                    module_name, init
                ).load_module()
                click.secho(f"\x1b[A[X] loaded {app_name}       ", fg="green")
            cls.modules[app_name] = module
            cls.errors[app_name] = None
        except Exception as err:
            cls.errors[app_name] = traceback.format_exc()
            error_logger.log(
                app_name, get_error_snapshot()
            )
            click.secho(
                f"\x1b[A[FAILED] loading {app_name} ({err})",
                fg="red",
            )
            # clear all files/submodules if the loading fails
            _clear_modules(module_name)
            return None

    @staticmethod
    def get_registered_routes():
        routes = []
        for route in om_app.routes.values():
            for method, method_obj in route.methods.items():
                func = method_obj.handler
                rule = route.rule
                routes.append(
                    {
                        "rule": rule,
                        "method": method,
                        "filename": module2filename(func.__module__),
                        "action": func.__name__,
                    }
                )
        return [*sorted(routes, key=lambda item: item["rule"])]



import ombott
import logging

from . import globs
from .core_events import core_event_bus, CoreEvents
from ..error_pages import error_page
from .install import install_args
from .reloader import Reloader
from .loggers import get_error_snapshot, error_logger
from .exceptions import HTTP


__all__ = (
    'import_apps',
    'reload_apps',
    'wsgi',
)


def import_apps():
    core_event_bus.emit(CoreEvents.BEFORE_APPS_LOAD)
    Reloader.import_apps()
    core_event_bus.emit(CoreEvents.APPS_LOADED)


@core_event_bus.on(CoreEvents.RELOAD_APPS)
def reload_apps(*app_names):
    Reloader.reimport_apps(*app_names)
    core_event_bus.emit(CoreEvents.APPS_LOADED)


def wsgi(**kwargs):
    """Initializes everything, loads apps, returns the wsgi app"""
    install_args(kwargs)
    import_apps()
    # ICECUBE.update(threadsafevariable.ThreadSafeVariable.freeze())
    return globs.app




"""WEBSAW - a web framework for rapid development with pleasure"""

import os
import code
import click
import sys
import platform
import json
import pydal
import zipfile
import signal
import asyncio
import logging
import ombott
import inspect

from .watcher import watch

# Optional web servers for speed
try:
    import gunicorn
except ImportError:
    gunicorn = None

from .core.install import install_args
from .core import import_apps
from . import server_adapters


WEBSAW_CMD = sys.argv[0]

ART = r"""
                   __
  ___     ______  / /_  _________ __      __
  | | /| / /> _ \/ __ \/ ___/ __ `/ | /| / />
  | |/ |/ />  __/ /_/ (__  ) /_/ /| |/ |/ />
  |__/|__/>\___/_.___/____/\__,_/ |__/|__/>

it is just the beginning...
"""

#########################################################################################
# CLI
#########################################################################################

__ssl__ = __import__("ssl")
_ssl = getattr(__ssl__, "_ssl") or getattr(__ssl__, "_ssl2")


def keyboardInterruptHandler(signal, frame):
    """Catch interrupts like Ctrl-C"""
    click.echo(
        "KeyboardInterrupt (ID: {}) has been caught. Cleaning up...".format(signal)
    )
    sys.exit(0)


@click.group(
    context_settings=dict(help_option_names=["-h", "-help", "--help"]),
    help='%s\n\nType "%s COMMAND -h" for available options on commands'
    % (__doc__, WEBSAW_CMD),
)
def cli():
    pass


@cli.command()
@click.option(
    "-a", "--all", is_flag=True, default=False, help="List version of all modules"
)
def version(all):
    """Show versions and exit"""
    from . import __version__

    click.echo("websaw: %s" % __version__)
    if all:
        click.echo("system: %s" % platform.platform())
        click.echo("python: %s" % sys.version.replace("\n", " "))
        for name in sorted(sys.modules):
            if hasattr(sys.modules[name], "__version__"):
                click.echo("%s: %s" % (name, sys.modules[name].__version__))


@cli.command()
@click.argument("apps_folder")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
def setup(**kwargs):
    """Setup new apps folder or reinstall it"""
    install_args(kwargs, reinstall_apps=True)


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
def shell(**kwargs):
    """Open a python shell with apps_folder's parent added to the path"""
    install_args(kwargs)
    code.interact(local=dict(globals(), **locals()))


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.argument("func")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option(
    "--args",
    default="{}",
    help="Arguments passed to the program/function",
    show_default=True,
)
def call(apps_folder, func, yes, args):
    """Call a function inside apps_folder"""
    kwargs = json.loads(args)
    install_args(dict(apps_folder=apps_folder, yes=yes))
    apps_folder_name = os.path.basename(os.environ["WEBSAW_APPS_FOLDER"])
    module, name = ("%s.%s" % (apps_folder_name, func)).rsplit(".", 1)
    env = {}
    exec("from %s import %s" % (module, name), {}, env)
    env[name](**kwargs)


@cli.command(name="set_password")
@click.option(
    "--password",
    prompt=True,
    confirmation_prompt=True,
    hide_input=True,
    help="Password value (asked if missing)",
)
@click.option(
    "-p",
    "--password_file",
    default="password.txt",
    help="File for the encrypted password",
    show_default=True,
)
def set_password(password, password_file):
    """Set administrator's password for the Dashboard"""
    click.echo('Storing the hashed password in file "%s"\n' % password_file)
    with open(password_file, "w") as fp:
        fp.write(str(pydal.validators.CRYPT()(password)[0]))


@cli.command(name="new_app")
@click.argument("apps_folder")
@click.argument("app_name")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option(
    "-s",
    "--scaffold_zip",
    default=None,
    help="Path to the zip with the scaffolding app",
    show_default=False,
)
def new_app(apps_folder, app_name, yes, scaffold_zip):
    """Create a new app copying the scaffolding one"""
    install_args(dict(apps_folder=apps_folder, yes=yes))
    source = scaffold_zip or os.path.join(
        os.path.dirname(__file__), "assets", "websaw.app._scaffold.zip"
    )
    target_dir = os.path.join(os.environ["WEBSAW_APPS_FOLDER"], app_name)
    if not os.path.exists(source):
        click.echo("Source app %s does not exists" % source)
        sys.exit(1)
    elif os.path.exists(target_dir):
        click.echo("Target folder %s already exists" % target_dir)
        sys.exit(1)
    else:
        zfile = zipfile.ZipFile(source, "r")
        zfile.extractall(target_dir)
        zfile.close()


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option("-H", "--host", default="127.0.0.1", help="Host name", show_default=True)
@click.option(
    "-P", "--port", default=8000, type=int, help="Port number", show_default=True
)
@click.option(
    "-p",
    "--password_file",
    default="password.txt",
    help="File for the encrypted password",
    show_default=True,
)
@click.option(
    "-s",
    "--server",
    default="default",
    type=click.Choice(
        ["default", "wsgiref", "gunicorn", "gevent", "waitress"]
        + server_adapters.__all__
    ),
    help="server to use",
    show_default=True,
)
@click.option(
    "-w",
    "--number_workers",
    default=0,
    type=int,
    help="Number of workers",
    show_default=True,
)
@click.option(
    "-d",
    "--dashboard_mode",
    default="full",
    help="Dashboard mode: demo, readonly, full, none",
    show_default=True,
)
@click.option(
    "--watch",
    default="lazy",
    type=click.Choice(["off", "sync", "lazy"]),
    help="Watch python changes and reload apps automatically, modes: off, sync, lazy",
    show_default=True,
)
@click.option(
    "--ssl_cert", type=click.Path(exists=True), help="SSL certificate file for HTTPS"
)
@click.option("--ssl_key", type=click.Path(exists=True), help="SSL key file for HTTPS")
def run(**kwargs):
    """Run all the applications on apps_folder"""
    install_args(kwargs)

    from . import __version__

    click.secho(ART, fg="cyan")
    click.echo("Websaw: %s on Python %s\n\n" % (__version__, sys.version))

    # If we know where the password is stored, read it, otherwise ask for one
    if os.path.exists(os.path.join(os.environ["WEBSAW_APPS_FOLDER"], "_dashboard")):
        if (
            kwargs["dashboard_mode"] not in ("demo", "none")
            and not os.path.exists(kwargs["password_file"])
        ):
            click.echo(
                'You have not set a dashboard password. Run "%s set_password" to do so.'
                % WEBSAW_CMD
            )
        else:
            click.echo(
                "Dashboard is at: http://%s:%s/_dashboard"
                % (kwargs["host"], kwargs["port"])
            )

    # Start
    import_apps()
    start_server(kwargs)


def start_server(kwargs):
    host = kwargs["host"]
    port = int(kwargs["port"])
    apps_folder = kwargs["apps_folder"]
    number_workers = kwargs["number_workers"]
    params = dict(host=host, port=port, reloader=False)
    server_config = dict(
        platform=platform.system().lower(),
        server=None if kwargs["server"] == "default" else kwargs["server"],
        number_workers=number_workers,
    )

    if not server_config["server"]:
        if server_config["platform"] == "windows" or number_workers < 2:
            server_config["server"] = "rocket"
        else:
            if not gunicorn:
                logging.error("gunicorn not installed")
                return
            server_config["server"] = "gunicorn"

    # Catch interrupts like Ctrl-C if needed
    if server_config["server"] not in {"rocket", "wsgirefWsTwistedServer"}:
        signal.signal(signal.SIGINT, keyboardInterruptHandler)

    params["server"] = server_config["server"]
    if params["server"] in server_adapters.__all__:
        params["server"] = getattr(server_adapters, params["server"])()
    if number_workers > 1:
        params["workers"] = number_workers
    if server_config["server"] == "gunicorn":
        sys.argv[:] = sys.argv[:1]  # else break gunicorn
    if kwargs["ssl_cert"] is not None:
        params["certfile"] = kwargs["ssl_cert"]
        params["keyfile"] = kwargs["ssl_key"]

    if server_config["server"] == "gevent":
        if not hasattr(_ssl, "sslwrap"):
            _ssl.sslwrap = new_sslwrap

    if kwargs["watch"] != "off":
        watch(apps_folder, server_config, kwargs["watch"])
    ombott.run(**params)


def new_sslwrap(
    sock,
    server_side=False,
    keyfile=None,
    certfile=None,
    cert_reqs=__ssl__.CERT_NONE,
    ssl_version=__ssl__.PROTOCOL_SSLv23,
    ca_certs=None,
    ciphers=None,
):
    context = __ssl__.SSLContext(ssl_version)
    context.verify_mode = cert_reqs or __ssl__.CERT_NONE
    if ca_certs:
        context.load_verify_locations(ca_certs)
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    if ciphers:
        context.set_ciphers(ciphers)
    caller_self = inspect.currentframe().f_back.f_locals["self"]
    return context._wrap_socket(sock, server_side=server_side, ssl_sock=caller_self)



import renoir
import renoir.constants
import renoir.writers

from .utils import safely, Cache


class RenoirXMLEscapeMixin:
    def _escape_data(self, data):
        """Allows Renoir to convert yatl helpers to strings"""
        return safely(
            lambda: data.xml(), default=lambda: self._to_html(self._to_unicode(data))
        )


class RenoirCustomWriter(RenoirXMLEscapeMixin, renoir.writers.Writer):
    ...


class RenoirCustomEscapeAllWriter(RenoirXMLEscapeMixin, renoir.writers.EscapeAllWriter):
    ...


class Renoir(renoir.Renoir):
    """Custom Renoir Engine that understands yatl helpers"""
    _writers = {
        renoir.constants.ESCAPES.common: RenoirCustomWriter,
        renoir.constants.ESCAPES.all: RenoirCustomEscapeAllWriter,
    }


def render(
    content=None,
    filename=None,
    path=".",
    context={},
    delimiters="[[ ]]",
    cached_renoir_engines=Cache(100),
):
    """
    renders the template using renoire, same API as yatl.render, does caching of
    both Renoire engine and source files
    """
    engine = cached_renoir_engines.get(
        (path, delimiters),
        lambda: Renoir(path=path, delimiters=delimiters.split(" "), reload=True),
    )
    if content is not None:
        return engine._render(content, context=context)
    return engine.render(filename, context=context)



import inspect
import pathlib
import os
import logging
import click
import asyncio
import traceback
import threading
from watchgod import awatch

from .core.core_events import core_event_bus, CoreEvents
from .core.globs import request_hooks


DIRTY_APPS = set()  # apps that need to be reloaded (lazy watching)


APP_WATCH = {"files": dict(), "handlers": dict(), "tasks": dict()}

""" Decorator that binds a func as an watchdog handler of non-'.py' files.
Paths to files must be relative to app, w/o app name(folder).

@app_watch_handler(['static/sass/all.sass', 'static/sass/main.sass'])
def sass_compile(changed_files):
    print(changed_files); # paths of files that changed, for info
    sass.compile()
"""


def app_watch_handler(watched_app_subpaths):
    stack = inspect.stack
    invoker = pathlib.Path(stack()[1].filename)
    apps_path = pathlib.Path(os.environ["WEBSAW_APPS_FOLDER"])
    app = invoker.relative_to(os.environ["WEBSAW_APPS_FOLDER"]).parts[0]

    def decorator(func):
        handler = "{}.{}".format(func.__module__, func.__name__)
        APP_WATCH["handlers"][handler] = func
        for subpath in watched_app_subpaths:
            app_path = apps_path.joinpath(app, subpath).as_posix()
            if app_path not in APP_WATCH["files"]:
                APP_WATCH["files"][app_path] = []
            APP_WATCH["files"][app_path].append(handler)
        return func

    return decorator


def try_app_watch_tasks():
    if APP_WATCH["tasks"]:
        tried_tasks = []
        for handler in APP_WATCH["tasks"]:
            changed_files_dict = APP_WATCH["tasks"][handler]
            try:
                APP_WATCH["handlers"][handler](changed_files_dict.keys())
                tried_tasks.append(handler)
            except Exception:
                logging.error(traceback.format_exc())
        ## remove executed tasks from register
        for handler in tried_tasks:
            del APP_WATCH["tasks"][handler]


def watch(apps_folder, server_config, mode="sync"):
    def watch_folder_event_loop(apps_folder):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(watch_folder(apps_folder))

    async def watch_folder(apps_folder):
        click.echo(
            "watching (%s-mode) python file changes in: %s" % (mode, apps_folder)
        )
        try:
            async for changes in awatch(os.path.join(apps_folder)):
                apps = []
                for subpath in [pathlib.Path(pair[1]) for pair in changes]:
                    name = subpath.relative_to(apps_folder).parts[0]
                    if subpath.suffix == ".py":
                        apps.append(name)
                    ## manage `app_watch_handler` decorators
                    elif subpath.as_posix() in APP_WATCH["files"]:
                        handlers = APP_WATCH["files"][subpath.as_posix()]
                        for handler in handlers:
                            if handler not in APP_WATCH["tasks"]:
                                APP_WATCH["tasks"][handler] = {}
                            APP_WATCH["tasks"][handler][subpath.as_posix()] = True

                if mode == "lazy":
                    DIRTY_APPS.update(apps)
                else:
                    core_event_bus.emit(CoreEvents.RELOAD_APPS, *apps)
                    try_app_watch_tasks()
        except RuntimeError as exc:
            if str(exc).endswith('after shutdown'):
                pass

    if server_config["number_workers"] > 1:
        click.echo("--watch option has no effect in multi-process environment \n")
        return
    elif server_config["server"].startswith(("wsgiref", "waitress", "rocket")):
        # these servers block the main thread so we open a new thread for the file watcher
        threading.Thread(
            target=watch_folder_event_loop, args=(apps_folder,), daemon=True
        ).start()
    elif server_config["server"].startswith("gevent"):
        watch_folder_event_loop(apps_folder)
    else:
        # should never happen
        return

    if mode == "lazy":
        request_hooks.before.add(lazy_trigger)


def lazy_trigger(*a, **kw):
    core_event_bus.emit(CoreEvents.RELOAD_APPS, *DIRTY_APPS)
    DIRTY_APPS.clear()
    try_app_watch_tasks()



import sys
import os
import json
import ombott
import logging

from websaw.core.loggers import get_error_snapshot, error_logger
from .error_pages import error_page

from .core import (
    request,
    response,
    abort,

    import_apps,
    reload_apps,
    wsgi,
    BaseApp,
    BaseContext,
    Reloader,

    redirect,
    Cache,

    HTTP,
    WebsawException,
)

from .fixtures import (
    Template,
    Session, GroupSession,
    DAL,
    URL,
    XAuth,
    AuthGuard,
    AuthErr,
    CurrentUser,
)

from pydal import Field

__all__ = (
    'request',
    'response',
    'abort',

    'import_apps',
    'reload_apps',
    'wsgi',
    'BaseApp',
    'BaseContext',

    'URL',
    'redirect',
    'Cache',

    'Template',
    'Session', 'GroupSession',
    'DAL',
    'Field',

    'HTTP',
    'WebsawException',

    'XAuth',
    'AuthGuard',
    'AuthErr',
    'CurrentUser',
)

__author__ = "Kucherov Valery <valq7711@gmail.com>"
__license__ = "MIT"
__version__ = "0.0.7"


def _maybe_gevent():
    for arg in sys.argv[1:]:
        if 'gevent' in arg.lower():
            from gevent import monkey
            monkey.patch_all()
            break
_maybe_gevent()  # noqa


class DefaultContext(BaseContext):
    group_session = GroupSession()
    session = Session()
    URL = URL()
    auth_guard = AuthGuard()
    current_user = CurrentUser()


class DefaultApp(BaseApp):
    def __init__(self, ctx: BaseContext, config=None, *, name=None):
        if name is None:
            if self.__class__ is DefaultApp:
                raise ValueError(
                    f"{self.__class__} can't be directly instantiated without 'name' argument"
                )
            name = self.__module__
        name_split = name.split('.')
        app_name = name_split[-1]

        pjoin = os.path.join
        folder = pjoin(Reloader.get_apps_folder(), *name_split[1:])
        static_folder = pjoin(folder, 'static')
        template_folder = pjoin(folder, 'templates')

        cfg = dict(
            app_name=app_name,
            base_url=f'/{app_name}',
            static_base_url=f'/{app_name}',
            folder=folder,
            static_folder=static_folder,
            template_folder=template_folder,
            render_map={dict: jsonfy},
            exception_handler=_default_exception_handler,
            group_name=None,
        )
        if config:
            cfg.update(config)

        super().__init__(cfg, ctx)

    def use(self, *fixt):
        app_name = self.default_config['app_name']
        return super().use(*[
            Template(f, path=self.default_config['template_folder'], inject={'mixin_name': app_name})
            if isinstance(f, str) else f
            for f in fixt
        ])


def jsonfy(ctx: BaseContext, dct):
    ctx.response.headers['Content-Type'] = 'application/json'
    return json.dumps(dct, sort_keys=True, indent=2, ensure_ascii=False, default=str)


def _default_exception_handler(ctx: BaseContext, exc):
    response = ctx.response
    try:
        raise exc
    except HTTP as http:
        response.status = http.status
        ret = getattr(http, "body", "")
        http_headers = getattr(http, 'headers', None)
        if http_headers:
            response.headers.update(http_headers)
        ctx.output = ret
    except ombott.HTTPResponse as resp:
        ctx.output = resp
    except Exception:
        snapshot = get_error_snapshot()
        logging.error(snapshot["traceback"])
        ticket_uuid = error_logger.log(ctx.app_data.app_name, snapshot) or "unknown"
        ctx.output = ombott.HTTPResponse(
            body=error_page(
                500,
                button_text=ticket_uuid,
                href="/_dashboard/ticket/" + ticket_uuid,
            ),
            status=500,
        )



import os

import yatl

from ..core import Cache, render, BaseContext
from ..core.fixture import Fixture


_HELPERS = {name: getattr(yatl.helpers, name) for name in yatl.helpers.__all__}


class Template(Fixture):

    cache = Cache(100)

    def __init__(self, filename, path=None, delimiters="[[ ]]", inject=None):
        if ':' in filename:
            context_key, filename = filename.split(':')
        else:
            context_key = None
        self.context_key = context_key
        self.filename = filename
        self.path = path
        self.delimiters = delimiters
        self.inject = inject or {}

    def take_off(self, ctx: BaseContext):
        output = ctx.output
        if not isinstance(output, dict):
            return
        shared_data = ctx.state.shared_data
        path_join = os.path.join

        context = dict(
            _HELPERS,
            request=ctx.request,
            URL=ctx.URL,
            app_get=ctx.get,
            mixin_get=ctx.mixin_get,
            __vars__=output,
            **self.inject,
            **shared_data.get("template_context", {}),
            **output,
        )
        path = self.path or ctx.app_data.template_folder
        filename = path_join(path, self.filename)
        if not os.path.exists(filename):
            generic_filename = path_join(path, "generic.html")
            if os.path.exists(generic_filename):
                filename = generic_filename
        ctx.output = render(
            filename=filename, path=path, context=context, delimiters=self.delimiters
        )




import urllib
import threading
from types import SimpleNamespace

from ..core import Fixture, BaseContext

url_quote = urllib.parse.quote


class URL(Fixture):
    def __init__(
        self,
        script_name="SCRIPT_NAME",
        domain_map="HTTP_X_WEBSAW_APPNAME"
    ):
        self.script_name = script_name
        self.domain_map = domain_map

    def take_on(self, ctx: BaseContext):
        env_get = ctx.request.environ.get
        # FIXME
        original_url = env_get("HTTP_ORIGIN") or ctx.request.url
        scheme, _, domain = original_url.split("/", 3)[:3]

        url_ctx = SimpleNamespace(
            script_name=env_get(self.script_name, "").rstrip("/"),
            domain_mapped=env_get(self.domain_map),
            static_version=ctx.get('static_version'),
            scheme=scheme,
            domain=domain,
            base_url=ctx['base_url'],
            static_url=ctx['static_base_url'],
            named_routes=ctx['named_routes'],
        )
        URLBuilder.setup(url_ctx)

    def __call__(
            self, *parts, vars=None, hash=None, scheme=False, signer=None,
            use_appname=None, static_version=None, use_signer=False
    ):
        return URLBuilder(
            *parts, vars=vars, hash=hash, scheme=scheme, signer=signer,
            use_appname=use_appname, static_version=static_version, use_signer=use_signer
        )


class URLBuilder:
    """
    Examples:
    URL('a','b',vars=dict(x=1),hash='y')       -> /{script_name?}/{app_name?}/a/b?x=1#y
    URL('a','b',vars=dict(x=1),scheme=None)    -> //{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),scheme=True)    -> http://{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),scheme='https') -> https://{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),use_appname=False) -> /{script_name?}/a/b?x=1
    URL(':home')  -> /{script_name?}/{app_name?}/path/to/home_route
    URL(':profile', dict(user='tom'))  -> /{script_name?}/{app_name?}/path/to/profile/tom
    """

    _local = threading.local()
    _local.ctx = None

    @classmethod
    def setup(cls, url_ctx: SimpleNamespace):
        cls._local.ctx = url_ctx

    __slots__ = (
        'parts', 'vars', 'hash', 'scheme', 'signer',
        'use_appname', 'static_version', 'app_name',
        'use_signer',
    )

    def __init__(
            self, *parts, vars=None, hash=None, scheme=False, signer=None,
            use_appname=None, static_version=None, use_signer=False
    ):

        self.parts = parts
        self.vars = vars
        self.hash = hash
        self.scheme = scheme
        self.signer = signer
        self.use_appname = use_appname
        self.static_version = static_version
        self.use_signer = use_signer

    def sign(self, signer=None):
        if signer is not None:
            self.signer = signer
        self.use_signer = True
        return self

    def to_str(self):
        return self.__str__()

    def __str__(self):
        ctx = self._local.ctx

        vars = self.vars
        hash = self.hash
        static_version = self.static_version
        use_appname = self.use_appname
        parts = self.parts
        scheme = self.scheme
        base_url = ctx.base_url[1:]  # remove '/'
        is_abs_url = False
        is_named_route = False

        broken_parts = []
        if parts:
            len_parts = len(parts)
            idx = 0
            if str(parts[idx]).startswith(':'):
                route_name = parts[0][1:]
                route_args = []
                route_kw = {}
                idx += 1
                if idx < len_parts and isinstance(parts[idx], (list, tuple)):
                    route_args = parts[idx]
                    idx += 1
                if idx < len_parts and isinstance(parts[idx], dict):
                    route_kw = parts[idx]
                    idx += 1
                route_url = ctx.named_routes[route_name].url(*route_args, **route_kw)
                broken_parts = [route_url]
                is_named_route = True
            else:
                [broken_parts.extend(str(part).rstrip("/").split("/")) for part in parts]
                if broken_parts[0] == "static":
                    if static_version != "":
                        if static_version is None:
                            static_version = ctx.static_version
                        if static_version:
                            broken_parts.insert(1, f"_{static_version}")
                    base_url = ctx.static_url[1:]
                else:
                    if broken_parts[0] == '':  # i.e. startswith '/'
                        is_abs_url = True

        prefix = (ctx.script_name or '') + '/'
        if is_abs_url:
            prefix = prefix.rstrip('/')
        else:
            if not is_named_route:
                broken_parts.insert(0, base_url)
            if use_appname is None:
                use_appname = not ctx.domain_mapped
            if not use_appname:
                app, _, broken_parts[0] = broken_parts[0].partition('/')

        url_buff = [prefix, "/".join(url_quote(p) for p in broken_parts)]
        # Signs the URL if required.  Copy vars into urlvars not to modify it.
        urlvars = dict(vars) if vars else {}
        if self.use_signer:
            # Note that we need to sign the non-urlencoded URL, since
            # at verification time, it will be already URLdecoded.
            self.signer.sign(prefix + "/".join(broken_parts), urlvars)
        if urlvars:
            url_buff.append("?" + "&".join(
                f"{k}={url_quote(str(v))}" for k, v in urlvars.items()
            ))
        if hash:
            url_buff.append(f"#{hash}")

        scheme_domain = ''
        if scheme is not False:
            if scheme is True:
                scheme = ctx.scheme
            elif scheme is None:
                scheme = ""
            else:
                scheme += ':'
            scheme_domain = f'{scheme}//{ctx.domain}'
        return f'{scheme_domain}{"".join(url_buff)}'



import ombott
from ombott import SimpleConfig
from types import SimpleNamespace

ombott.DefaultConfig.max_memfile_size = 16 * 1024 * 1024

app = ombott.default_app()
app.setup()

request = app.request
response = app.response
static_file = ombott.static_file

request_hooks = SimpleNamespace(before=set())


def _before_request(*args, **kw):
    [h(*args, **kw) for h in request_hooks.before]


app.add_hook("before_request", _before_request)


@SimpleConfig.keys_holder
class DefaultConfig(SimpleConfig):
    apps_folder = 'apps'
    service_folder = ".service"
    service_db_uri = "sqlite://service.storage"
    password_file = 'password.txt'
    session_secret = None

    host = '127.0.0.1'
    port = 8000
    server = 'default'
    ssl_cert = None
    number_workers = 0

    dashboard_mode = 'full'
    watch = 'lazy'


# patched by install.py
current_config: DefaultConfig = SimpleNamespace()



import os
from types import SimpleNamespace


class ConfigSetter:
    def __init__(self, keys_class, defaults_factory=None, novalue=None):
        self.novalue = novalue
        self.defaults_factory = defaults_factory
        self.keys = {
            k for k in keys_class.__dict__
            if not k.startswith('_')
        }

    def extend(self, dst, src, optional_keys=None, get_default=None):
        self.apply(dst, src, optional_keys, get_default, extend_mode=False)

    def apply(self, dst, src, optional_keys=None, get_default=None, extend_mode=False):
        optional_keys = optional_keys or set()
        if get_default is None:
            if self.defaults_factory:
                get_default = self.defaults_factory(src).get
            else:
                get_default = {}.get

        no_value = self.novalue
        if isinstance(src, dict):
            get = src.get
        else:
            def get(k, default=None):
                return getattr(src, k, default)

        for k in self.keys:
            v = get(k, no_value)
            if extend_mode and v is not no_value:
                continue
            if v is no_value:
                try:
                    v = get_default(k)
                except KeyError:
                    if k in optional_keys:
                        continue
                    raise KeyError(f'Missing key: {k}')

            setattr(dst, k, v)

    def dict_from(self, src, get_default=None):
        defaults = get_default or {}.get
        ret = {}
        no_value = self.novalue
        for k in self.keys:
            v = getattr(src, k, no_value)
            if v is no_value:
                v = defaults(k, no_value)
            if v is no_value:
                raise KeyError(f'Missing key: {k}')
            ret[k] = v


class AppConfigKeys:
    folder = None

    # optional
    base_url = None
    static_base_url = None
    static_version = None

    static_folder = None
    template_folder = None


class AppConfig(AppConfigKeys):
    def __getitem__(s, k):
        return getattr(s, k)

    def __setitem__(s, k, v):
        return setattr(s, k, v)

    def get(s, k, d=None):
        return getattr(s, k, d)


class AppConfigDefaultsFactory:
    base_url = None
    static_version = '0.0.1'

    def __init__(self, src):
        if isinstance(src, dict):
            src = SimpleNamespace(**src)
        self.src = src

    def static_base_url(self):
        ret = getattr(self.src, 'base_url', None)
        return ret

    def folder(self):
        # to allow to use for static/template folders
        folder = getattr(self.src, 'folder', None)
        if folder is None:
            folder = '.'
        return folder

    def static_folder(self):
        return os.path.join(self.folder(), 'static')

    def template_folder(self):
        return os.path.join(self.folder(), 'templates')

    def get(self, k):
        no_value = []
        attr = getattr(self, k, no_value)
        if attr is no_value:
            raise KeyError(f'{k}')
        return attr() if callable(attr) else attr


_config_setter = ConfigSetter(AppConfigKeys, AppConfigDefaultsFactory)




import threading
from types import SimpleNamespace
from typing import Any, Optional, List, Dict

from .exceptions import FixtureProcessError
from . import globs
from .fixture import Fixture


class IContextSate:

    in_use: Dict[str, Fixture]
    in_use_stack: List[Fixture]
    output: Any
    exception_stack: List[Exception]
    shared_data: dict


class Local(threading.local):
    state: IContextSate
    in_process: bool
    mixin_data: SimpleNamespace


class MetaContext(type):

    def __new__(cls, name, bases, dct):
        if not bases:
            return super().__new__(cls, name, bases, dct)

        new_fixt = {
            k: dct.pop(k) for k, v in {**dct}.items()
            if isinstance(v, Fixture)
        }
        base_fixt = {}
        [
            base_fixt.update(getattr(bcls, '_fixt', {}))
            for bcls in reversed(bases)
        ]
        dct['_fixt'] = {**base_fixt, **new_fixt}
        out_cls = super().__new__(cls, name, bases, dct)
        return out_cls


class BaseContext(metaclass=MetaContext):

    __local__ = threading.local()

    _fixt: Dict[str, Fixture] = {}  # see metaclass

    _fixture_initialize = Fixture.initialize_safe_storage
    _fixture_prepare_for_use = Fixture.prepare_for_use

    request = globs.request
    response = globs.response
    app_data: Optional[SimpleNamespace] = None

    @classmethod
    def cctx(cls) -> 'BaseContext':
        return cls.__local__.current_ctx

    def __init__(self):
        self._fixt = {**self._fixt}
        self._reverse_map = {f: k for k, f in self._fixt.items()}
        self._local = Local()
        self._local.in_process = False
        self._local.mixin_data = None

    @property
    def state(self) -> IContextSate:
        return self._local.state

    @property
    def exception(self) -> Optional[Exception]:
        try:
            return self._local.state.exception_stack[0]
        except IndexError:
            pass

    @property
    def output(self):
        return self._local.state.output

    @output.setter
    def output(self, output):
        self._local.state.output = output

    @property
    def mixin_data(self):
        return self._local.mixin_data

    @mixin_data.setter
    def mixin_data(self, v):
        self._local.mixin_data = v

    def ask(self, fixture_key, default=None):
        return getattr(self, fixture_key, default)

    def get(self, app_attr, default=None):
        return getattr(self.app_data, app_attr, default)

    def mixin_get(self, mixin_attr, default=None):
        return getattr(self.mixin_data, mixin_attr, default)

    def __getitem__(self, k):
        return getattr(self.app_data, k)

    def __getattr__(self, k):
        local = self._local
        if not local.in_process:
            return self._fixt[k]
        state = local.state
        in_use = state.in_use
        f = in_use.get(k)
        if f is None:
            try:
                f = self._fixt[k]
            except KeyError:
                raise AttributeError(f'{k}')
            self._fixture_prepare_for_use(f)
            try:
                f.take_on(self)
            except Exception as exc:
                state.exception_stack.append(exc)
                raise FixtureProcessError()
            in_use[k] = f
            state.in_use_stack.append(f)
        return f

    def get_or_make_fixture_key(self, f) -> str:
        key = self._reverse_map.get(f)
        if key is None:
            key = getattr(f, 'context_key', None)
            if key is None:
                key = f'ID_{id(f)}'
            self._fixt[key] = f
            self._reverse_map[f] = key
        return key

    def clone(self, app_data=None):
        ret = self.__class__()
        ret._fixt = {**self._fixt}
        ret.app_data = app_data
        return ret

    def extend(self, *mixins):
        fixt = {}
        [fixt.update(m._fixt) for m in reversed(mixins)]
        [self._fixt.setdefault(k, f) for k, f in fixt.items()]

    def app_mounted(self):
        [f.app_mounted(self) for f in self._fixt.values()]

    def initialize(self):
        self.__local__.current_ctx = self
        local = self._local
        local.state = SimpleNamespace(
            in_use={},
            in_use_stack=[],
            output=None,
            exception_stack=[],
            shared_data={}
        )
        local.in_process = True
        self._fixture_initialize()

    def use_fixtures(self, fixt, hooks=False):
        try:
            [getattr(self, f) for f in fixt]
        except Exception:
            if hooks:
                self._local.state.in_use.update(hooks)
            raise

    def finalize(self, exc):
        state = self._local.state
        if exc:
            state.exception_stack.append(exc)

        max_cnt = 100
        in_use_stack = state.in_use_stack
        err = None
        while max_cnt > 0:
            max_cnt -= 1
            if not in_use_stack:
                break
            f = in_use_stack.pop()
            try:
                f.take_off(self)
            except Exception as exc:
                state.exception_stack.append(exc)
        else:
            err = RuntimeError('Max number of iterations exceeded')
        self._local.in_process = False
        if err:
            raise err




class WebsawException(Exception):
    pass


class FixtureProcessError(WebsawException):
    pass


class HTTP(WebsawException):

    def __init__(self, status, body=None):
        self.status = status
        self.body = body



import json
import calendar
import time
import uuid
from enum import Enum
from typing import Dict, Callable, Tuple

from ..core import Fixture, BaseContext, HTTP

from pydal.validators import (
    CRYPT,
)


class AuthErrEnumMixin(str):

    def __new__(cls, msg, code):
        mem = super().__new__(cls, msg)
        mem._value_ = msg
        mem.code = code
        return mem

    def __int__(self):
        return self.code

    def as_dict(self):
        return dict(code=self.code, msg=self.value)


class AuthErr(AuthErrEnumMixin, Enum):
    UNAUTHORIZED = ('unauthorized', 401)
    CREDENTIALS = ('credentials', 401)
    BANNED = ('banned', 403)
    TIMEOUT = ('timeout', 440)
    FORBIDDEN = ('forbidden', 403)


class CurrentUser(Fixture):

    def take_on(self, ctx: BaseContext):
        group_name = getattr(ctx.app_data, 'group_name', None)
        if group_name is not None:
            session: Fixture = ctx.group_session
        else:
            session: Fixture = ctx.session

        self.data.session = session

        user = session.get("user")
        user_id = user and user.get('id')

        if not user_id:
            if user is not None:
                self.clear()

    def take_off(self, ctx):
        if self.user:
            activity = self.recent_activity
            time_now = calendar.timegm(time.gmtime())
            if not activity or time_now - activity > 6:
                self.recent_activity = time_now
        else:
            self.recent_activity = None

    @property
    def recent_activity(self):
        return self.data.session.get("recent_activity")

    @recent_activity.setter
    def recent_activity(self, v):
        self.data.session["recent_activity"] = v
        return v

    @property
    def user(self):
        return self.data.session.get("user")

    def store_user_in_session(self, user: dict):
        session = self.data.session
        session["user"] = user
        session["uuid"] = str(uuid.uuid1())

    def clear(self):
        del self.data.session["user"]


class AuthGuard(Fixture):
    config_fkey = 'auth_guard_config'

    def __init__(
        self,
        session_timeout=3600,  # seconds,
        condition=None,
        error_handlers: Dict[AuthErr, Callable] = None
    ):

        if error_handlers is None:
            error_handlers = {}
        self.error_handlers = error_handlers
        self.condition = condition
        self.session_timeout = session_timeout

    def take_on(self, ctx: BaseContext):
        cuser: CurrentUser = ctx.current_user
        config: dict = ctx.ask(self.config_fkey, {})
        condition = config.get('condition', self.condition)
        user = cuser.user
        if cuser.user is None:
            cuser.recent_activity = None
            raise self._error(AuthErr.UNAUTHORIZED)

        activity = cuser.recent_activity
        time_now = calendar.timegm(time.gmtime())
        session_timeout = self.session_timeout
        # enforce the optionl auth session expiration time
        if session_timeout and activity:
            if time_now - activity > session_timeout:
                cuser.clear()
                raise self._error(AuthErr.TIMEOUT)
        # record the time of the latest activity for logged in user (with throttling)
        if condition and not condition(user):
            raise self._error(AuthErr.FORBIDDEN)
        self.data.user = user

    @property
    def user(self):
        return self.data.user

    def _error(self, err_type: AuthErr):
        cb = self.error_handlers.get(err_type)
        if cb:
            return cb()
        return HTTP(int(err_type), body=err_type.as_dict())


class XAuth(Fixture):

    _crypt = CRYPT()

    @classmethod
    def crypt(cls, pw: str):
        return cls._crypt(pw)[0]

    def __init__(self, id_field='id', ban_field='is_blocked'):
        self.id_field = id_field
        self.ban_field = ban_field

    def take_on(self, ctx: BaseContext):
        self.data.ctx = ctx
        self.data.cuser = ctx.current_user

    def user_by_login(self, login: str) -> dict:
        """Return user record with credentials.

        user: {'<self.id>', 'password', '<self.ban_field>'}
        ban-field is optional if `self.ban_field` is not set
        """
        raise NotImplementedError()

    def register(self, fields):
        raise NotImplementedError()

    def user_for_session(self, user):
        return {"id": user[self.id_field]}

    def store_user_in_session(self, user: dict):
        self.data.cuser.store_user_in_session(user)

    @property
    def user(self):
        return self.data.cuser.user

    @property
    def user_id(self):
        user = self.user
        return user and user["id"]

    @property
    def is_logged_in(self):
        return self.user is not None

    # Methods that do not assume a user

    def check_credentials(self, login: str, password: str) -> Tuple[dict, AuthErr]:
        user = self.user_by_login(login)
        if not user:
            return (None, AuthErr.CREDENTIALS)
        if self.ban_field and user[self.ban_field]:
            return (None, AuthErr.BANNED)
        crypt_pw = self.crypt(password)
        if crypt_pw != user['password']:
            return (None, AuthErr.CREDENTIALS)
        return (user, None)

    def login(self, name, password) -> Tuple[dict, AuthErr]:
        user, error = self.check_credentials(name, password)
        if user:
            self.store_user_in_session(self.user_for_session(user))
            return (user, None)
        return None, error

    def logout(self):
        self.data.cuser.clear()



import sys
import os
import click
import uuid
import zipfile

from .globs import current_config, DefaultConfig
from .loggers import error_logger
from .utils import MetaPathRouter


_ASSETS_DIR = os.path.normpath(
    os.path.join(os.path.dirname(__file__), "../assets")
)


def install_args(kwargs, reinstall_apps=False):

    # in fact, `config` is instance of ombott.common_helpers.NameSpace
    config: DefaultConfig = DefaultConfig(kwargs)

    apps_folder = config.apps_folder = os.path.abspath(config.apps_folder)
    config.service_folder = os.path.join(apps_folder, config.service_folder)

    for key, val in config.items():
        os.environ[f"WEBSAW_{key.upper()}"] = str(val)

    yes2 = yes = kwargs.get("yes", False)

    # If the apps folder does not exist create it and populate it
    if not os.path.exists(apps_folder):
        if yes or click.confirm(f"Create missing folder {apps_folder}?"):
            os.makedirs(apps_folder)
            yes2 = True
        else:
            click.echo("Command aborted")
            sys.exit(0)

    init_py = os.path.join(apps_folder, "__init__.py")
    if not os.path.exists(init_py):
        if yes2 or click.confirm(f"Create missing init file {init_py}?"):
            with open(init_py, "wb"):
                pass
        else:
            click.echo("Command aborted")
            sys.exit(0)

    # ensure that "import apps.someapp" works
    apps_folder_parent, apps_folder_name = os.path.split(apps_folder)
    if apps_folder_parent not in sys.path:
        sys.path.insert(0, apps_folder_parent)
    if apps_folder_name != "apps":
        MetaPathRouter(apps_folder_name)

    if not os.path.exists(config.service_folder):
        os.mkdir(config.service_folder)
    session_secret_filename = os.path.join(
        config.service_folder, "session.secret"
    )
    if not os.path.exists(session_secret_filename):
        with open(session_secret_filename, "w") as fp:
            fp.write(str(uuid.uuid4()))

    with open(session_secret_filename) as fp:
        config.session_secret = fp.read()

    error_logger.initialize()
    if reinstall_apps:
        reinstall_apps(apps_folder, yes)

    current_config.__dict__.update(config.__dict__)


def reinstall_apps(apps_folder, confirmed):
    assets_dir = _ASSETS_DIR
    # Reinstall apps from zipped ones in assets
    if os.path.exists(assets_dir):
        apps = os.listdir(assets_dir)
        for filename in apps:
            zip_filename = os.path.join(assets_dir, filename)
            # These filenames do not necessarily exist if one has
            # downloaded from source and deleted them.
            app_name = filename.split(".")[-2]
            target_dir = os.path.join(apps_folder, app_name)
            if not os.path.exists(target_dir):
                if confirmed or click.confirm(f"Create app {app_name}?"):
                    click.echo(f"[ ] Unzipping app {filename}")
                    with zipfile.ZipFile(zip_filename, "r") as zip_file:
                        os.makedirs(target_dir)
                        zip_file.extractall(target_dir)
                        click.echo("\x1b[A[X]")



import threading
from types import SimpleNamespace
from typing import Optional


class Fixture:

    _local = threading.local()
    is_hook = False
    context_key: Optional[str]

    @classmethod
    def initialize_safe_storage(cls):
        cls._local.fixtures_data = {}

    @classmethod
    def prepare_for_use(cls, f):
        cls._local.fixtures_data[f] = SimpleNamespace()

    @property
    def data(self) -> SimpleNamespace:
        return self._local.fixtures_data[self]

    def app_mounted(self, ctx):
        ...

    def take_on(self, ctx):
        ...

    def take_off(self, ctx):
        ...
